/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/windows/thread.c: Thread management for Windows.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/windows/internal.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Structure to hold thread data.  A pointer to this structure is returned
 * as the thread handle.  (If not for CALLBACK, i.e. __stdcall, we could
 * just pass the function pointer directly to CreateThread().  Sigh...) */

typedef struct SysThread SysThread;
struct SysThread {
    /* Windows thread handle. */
    HANDLE handle;
    /* Windows thread ID, as returned by GetCurrentThreadId().  Set by the
     * thread runner.  (This is needed because GetThreadId() is missing
     * from Windows XP.) */
    DWORD id;

    /* Function to call, and its parameter. */
    int (*function)(void *param);
    void *param;
};

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * thread_runner:  Wrapper for threads started with sys_thread_create()
 * which handles translating between standard and Windows calling conventions.
 *
 * [Parameters]
 *     param: Thread parameter (points to a SysThread structure).
 * [Return value]
 *     Thread function return value.
 */
static CALLBACK DWORD thread_runner(LPVOID param);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysThreadID sys_thread_create(
    int priority, int stack_size, int (*function)(void *), void *param)
{
    if (priority != THREAD_PRIORITY_IDLE
     && priority != THREAD_PRIORITY_TIME_CRITICAL) {
        if (priority < THREAD_PRIORITY_LOWEST) {
            DLOG("Thread priority %d not supported, clamping to %d",
                 priority, THREAD_PRIORITY_LOWEST);
            priority = THREAD_PRIORITY_LOWEST;
        } else if (priority > THREAD_PRIORITY_HIGHEST) {
            DLOG("Thread priority %d not supported, clamping to %d",
                 priority, THREAD_PRIORITY_HIGHEST);
            priority = THREAD_PRIORITY_HIGHEST;
        }
    }

    SysThread *thread = mem_alloc(sizeof(*thread), 0, 0);
    if (!thread) {
        DLOG("No memory for thread structure");
        return 0;
    }
    thread->function = function;
    thread->param = param;
    /* Set thread->id to an invalid value to avoid spurious wait errors if
     * the caller tries to wait for the thread before the thread sets
     * thread->id.  Microsoft's documentation explicitly says that thread
     * IDs will never be zero. */
    thread->id = 0;
    thread->handle = CreateThread(NULL, stack_size, thread_runner, thread,
                                  CREATE_SUSPENDED, NULL);
    if (!thread->handle) {
        DLOG("Failed to create thread for %p(%p)", function, param);
        mem_free(thread);
        return 0;
    }
    if (!SetThreadPriority(thread->handle, priority)) {
        DLOG("Failed to set thread priority for %p(%p) (running anyway)",
             function, param);
    }
    if (!ResumeThread(thread->handle)) {
        DLOG("Failed to run thread for %p(%p)", function, param);
        CloseHandle(thread->handle);
        mem_free(thread);
        return 0;
    }
    return (SysThreadID)thread;
}

/*-----------------------------------------------------------------------*/

int sys_thread_is_running(SysThreadID thread_)
{
    SysThread *thread = (SysThread *)thread_;
    return WaitForSingleObject(thread->handle, 0) != WAIT_OBJECT_0;
}

/*-----------------------------------------------------------------------*/

int sys_thread_wait(SysThreadID thread_, int *result_ret)
{
    SysThread *thread = (SysThread *)thread_;

    /* WaitForSingleObject() won't protect against trying to wait for the
     * current thread, so we need to check manually whether the target
     * thread is the same as the current thread.  Note that while we don't
     * wait for thread->id to be set before returning from
     * sys_thread_create(), this condition can only be true when called
     * from the thread that set thread->id in the first place, so locking
     * isn't necessary to get correct behavior. */
    if (thread->id == GetCurrentThreadId()) {
        DLOG("Attempted to wait for current thread!");
        return 0;
    }

    if (WaitForSingleObject(thread->handle, INFINITE) == WAIT_FAILED) {
        DLOG("Failed to wait for thread: %s",
             windows_strerror(GetLastError()));
        return 0;
    }

    DWORD result = 0;
    if (!GetExitCodeThread(thread->handle, &result)) {
        DLOG("Failed to get thread exit code for thread %p", thread->handle);
    }
    CloseHandle(thread->handle);
    mem_free(thread);
    *result_ret = (int)result;
    return 1;
}

/*-----------------------------------------------------------------------*/

void sys_thread_yield()
{
    SwitchToThread();
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static CALLBACK DWORD thread_runner(LPVOID param)
{
    SysThread *thread = (SysThread *)param;
    thread->id = GetCurrentThreadId();
    int result = (*thread->function)(thread->param);
    return (DWORD)result;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
