/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/sdl/graphics.c: SDL-specific graphics-related tests.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/sysdep/sdl/dyngl.h"
#include "src/sysdep/sdl/internal.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"

#include <SDL.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Override vendor and renderer strings for override_glGetString(), used
 * to test blacklisting. */
static const char *override_vendor;
static const char *override_renderer;

/**
 * override_glGetString:  Override function for glGetString().  When name
 * is GL_VENDOR (resp. GL_RENDERER), this function returns the value of
 * override_vendor (resp. override_renderer) if it is non-NULL.
 */
static const GLubyte * GLAPIENTRY override_glGetString(
    GLenum name, const GLubyte * GLAPIENTRY (*original_glGetString)(GLenum));

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

static int do_test_sdl_graphics(void);
int test_sdl_graphics(void)
{
    return run_tests_in_window(do_test_sdl_graphics);
}

DEFINE_GENERIC_TEST_RUNNER(do_test_sdl_graphics)

TEST_INIT(init)
{
    graphics_start_frame();
    graphics_clear(0, 0, 0);
    return 1;
}

TEST_CLEANUP(cleanup)
{
    graphics_finish_frame();
    graphics_flush_resources();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_sdl_window)
{
#if SDL_VERSION_ATLEAST(2,0,0)
    CHECK_TRUE(sdl_window());
#else
    CHECK_FALSE(sdl_window());
#endif
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_dyngl_double_init)
{
    CHECK_TRUE(dyngl_init());
    CHECK_TRUE(glGetString(GL_VERSION));
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_dyngl_has_compressed_textures)
{
    if (opengl_major_version() >= 2
        || (opengl_major_version() == 1 && opengl_minor_version() >= 3)
        || opengl_has_extension("ARB_texture_compression"))
    {
        CHECK_TRUE(dyngl_has_compressed_textures());
    } else {
        CHECK_FALSE(dyngl_has_compressed_textures());
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_dyngl_has_vertex_buffers)
{
    if (opengl_major_version() >= 2
        || (opengl_major_version() == 1 && opengl_minor_version() >= 5)
        || opengl_has_extension("ARB_vertex_buffer_object"))
    {
        CHECK_TRUE(dyngl_has_vertex_buffers());
    } else {
        CHECK_FALSE(dyngl_has_vertex_buffers());
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_dyngl_has_shaders)
{
    if (opengl_major_version() >= 2) {
        CHECK_TRUE(dyngl_has_shaders());
    } else {
        CHECK_FALSE(dyngl_has_shaders());
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_dyngl_has_framebuffers)
{
    if (opengl_major_version() >= 3
        || opengl_has_extension("EXT_framebuffer_object"))
    {
        CHECK_TRUE(dyngl_has_framebuffers());
    } else {
        CHECK_FALSE(dyngl_has_framebuffers());
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_dyngl_has_opengl3)
{
    if (opengl_major_version() >= 3) {
        CHECK_TRUE(dyngl_has_opengl3());
    } else {
        CHECK_FALSE(dyngl_has_opengl3());
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_blacklist)
{
#ifdef SIL_PLATFORM_WINDOWS
    /* We can't run this test on Windows because initializing dyngl early
     * gives us only OpenGL 1.1. */
    SKIP("Can't run this test on Windows.");
#endif

    graphics_finish_frame();  // We don't do any rendering here.

    /* These values should fail to match any blacklist entry, so even if
     * the real renderer is blacklisted, we can run the test (assuming the
     * OpenGL implementation itself supports shaders and framebuffers). */
    override_vendor = "_";
    override_renderer = "_";
    graphics_cleanup();
    TEST_dyngl_override_glGetString(override_glGetString);
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    const int has_shaders = opengl_has_features(OPENGL_FEATURE_SHADERS);
    const int has_framebuffers =
        opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS);
    if (!has_shaders || !has_framebuffers) {
        SKIP("Renderer doesn't support both shaders and framebuffers.");
    }

    override_vendor = "__MATCH__";
    override_renderer = "__TEST__";
    graphics_cleanup();
    TEST_dyngl_override_glGetString(override_glGetString);
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    CHECK_TRUE(opengl_has_features(OPENGL_FEATURE_SHADERS));
    CHECK_FALSE(opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS));

    override_vendor = "__MATCH__";
    override_renderer = "__TST__";
    graphics_cleanup();
    TEST_dyngl_override_glGetString(override_glGetString);
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    CHECK_TRUE(opengl_has_features(OPENGL_FEATURE_SHADERS));
    CHECK_TRUE(opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS));

    override_vendor = "__TEST__";
    override_renderer = "__MATCH__";
    graphics_cleanup();
    TEST_dyngl_override_glGetString(override_glGetString);
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    CHECK_FALSE(opengl_has_features(OPENGL_FEATURE_SHADERS));
    CHECK_TRUE(opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS));

    override_vendor = "__TST__";
    override_renderer = "__MATCH__";
    graphics_cleanup();
    TEST_dyngl_override_glGetString(override_glGetString);
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    CHECK_TRUE(opengl_has_features(OPENGL_FEATURE_SHADERS));
    CHECK_TRUE(opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS));

    override_vendor = "__FOO__";
    override_renderer = "__REGEX__";
    graphics_cleanup();
    TEST_dyngl_override_glGetString(override_glGetString);
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    CHECK_FALSE(opengl_has_features(OPENGL_FEATURE_SHADERS));
    CHECK_FALSE(opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS));

    override_vendor = "__FOO__";
    override_renderer = "__RGX__";
    graphics_cleanup();
    TEST_dyngl_override_glGetString(override_glGetString);
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    CHECK_TRUE(opengl_has_features(OPENGL_FEATURE_SHADERS));
    CHECK_TRUE(opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS));

    override_vendor = "__FOO__";
    override_renderer = "__RGEX__";
    graphics_cleanup();
    TEST_dyngl_override_glGetString(override_glGetString);
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    CHECK_FALSE(opengl_has_features(OPENGL_FEATURE_SHADERS));
    CHECK_FALSE(opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS));

    /* Reset stuff so the cleanup routine doesn't crash. */
    graphics_cleanup();
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    graphics_start_frame();

    return 1;
}

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

static const GLubyte * GLAPIENTRY override_glGetString(
    GLenum name, const GLubyte * GLAPIENTRY (*original_glGetString)(GLenum))
{
    if (name == GL_VENDOR && override_vendor) {
        return (const GLubyte *)override_vendor;
    }
    if (name == GL_RENDERER && override_renderer) {
        return (const GLubyte *)override_renderer;
    }
    return (*original_glGetString)(name);
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
